from __future__ import annotations

import os
from pathlib import Path
from tempfile import gettempdir
from typing import Any, Dict

import click
from pydantic import BaseSettings, validator
from pydantic.utils import deep_update as _deep_update

from .utils import Literal


class BaseConfig(BaseSettings):
    def _build_values(self, init_kwargs: Dict[str, Any]) -> Dict[str, Any]:
        # Prioritise env vars
        return _deep_update(init_kwargs, self._build_environ())


class _Config(BaseConfig):
    config_dir: Path = None   # type: ignore  # https://github.com/samuelcolvin/pydantic/issues/866
    addon_dir: Path
    temp_dir: Path = Path(gettempdir()) / 'instawow'
    game_flavour: Literal['retail', 'classic']

    class Config:
        env_prefix = 'INSTAWOW_'

    @validator('config_dir', pre=True, always=True)
    def _apply_config_dir_default(cls, value: Any) -> Path:
        return Path(value or click.get_app_dir('instawow'))

    @validator('config_dir', 'addon_dir', 'temp_dir')
    def _expand_paths(cls, value: Path) -> Path:
        try:
            return Path(value).expanduser().resolve()
        except RuntimeError as error:
            # pathlib will raise RuntimeError for non-existent ~users
            raise ValueError(str(error)) from error

    @validator('addon_dir')
    def _check_writable(cls, value: Path) -> Path:
        if not (value.is_dir() and os.access(value, os.W_OK)):
            raise ValueError('must be a writable directory')
        return value

    @classmethod
    def read(cls) -> _Config:
        dummy_config = cls(addon_dir='', game_flavour='retail')
        return cls.parse_raw(dummy_config.config_file.read_text(encoding='utf-8'))

    def ensure_dirs(self) -> _Config:
        self.config_dir.mkdir(exist_ok=True, parents=True)
        for dir_ in self.logger_dir, self.plugin_dir, self.temp_dir, self.cache_dir:
            dir_.mkdir(exist_ok=True)
        return self

    def write(self) -> _Config:
        # This is separate from ``ensure_dirs`` because configuration
        # values are considered to be transient if overridden by an
        # env var, e.g. to bypass the cache
        self.ensure_dirs()
        output = self.json(exclude={'config_dir'}, indent=2)
        self.config_file.write_text(output, encoding='utf-8')
        return self

    @property
    def is_classic(self) -> bool:
        return self.game_flavour == 'classic'

    @property
    def is_retail(self) -> bool:
        return self.game_flavour == 'retail'

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'

    @property
    def logger_dir(self) -> Path:
        return self.config_dir / 'logs'

    @property
    def plugin_dir(self) -> Path:
        return self.config_dir / 'plugins'

    @property
    def cache_dir(self) -> Path:
        return self.temp_dir / '.cache'


Config = _Config



from __future__ import annotations

from datetime import datetime
from enum import Enum
from functools import wraps
from itertools import takewhile
import json
from operator import itemgetter
import re
from typing import (TYPE_CHECKING, cast,
                    Any, Callable, ClassVar, Dict, List, NamedTuple, Optional, Set, Tuple)

from yarl import URL

from . import exceptions as E, models as m
from .utils import (ManagerAttrAccessMixin, gather, is_not_stale, run_in_thread as t, shasum,
                    slugify)

if TYPE_CHECKING:
    from .manager import Manager


class Strategies(Enum):
    default = 'default'
    latest = 'most recent file available'
    curse_latest_beta = 'most recent beta quality file available from CurseForge'
    curse_latest_alpha = 'most recent alpha quality file available from CurseForge'


class Defn(NamedTuple):
    source: str
    name: str
    strategy: Strategies = Strategies.default

    def with_name(self, name: str) -> Defn:
        return self.__class__(self.source, name, self.strategy)

    def with_strategy(self, strategy: Strategies) -> Defn:
        return self.__class__(self.source, self.name, strategy)

    def __str__(self) -> str:
        return f'{self.source}:{self.name}'


def validate_strategy(method: Callable) -> Callable:
    @wraps(method)
    async def wrapper(self, defn: Defn, *args: Any, **kwargs: Any) -> m.Pkg:
        if defn.strategy in self.strategies:
            return await method(self, defn, *args, **kwargs)
        raise E.PkgStrategyUnsupported(defn.strategy)

    return wrapper


class _FileCacheMixin:
    async def _cache_json_response(self: Any, url: str, *args: Any) -> Any:
        path = self.config.temp_dir / shasum(url)

        if await t(is_not_stale)(path, *args):
            text = await t(path.read_text)(encoding='utf-8')
        else:
            async with self.web_client.get(url, raise_for_status=True) as response:
                text = await response.text()
            await t(path.write_text)(text, encoding='utf-8')
        return json.loads(text)


class Resolver(ManagerAttrAccessMixin):
    source: ClassVar[str]
    name: ClassVar[str]
    strategies: ClassVar[Set[Strategies]]

    def __init__(self, manager: Manager) -> None:
        self.manager = manager

    @classmethod
    def decompose_url(cls, value: str) -> Optional[Tuple[str, str]]:
        "Attempt to extract definition names from resolvable URLs."

    async def synchronise(self) -> Resolver:
        return self

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        raise NotImplementedError


class CurseResolver(Resolver, _FileCacheMixin):
    source = 'curse'
    name = 'CurseForge'
    strategies = {Strategies.default,
                  Strategies.latest,
                  Strategies.curse_latest_beta,
                  Strategies.curse_latest_alpha}

    # https://twitchappapi.docs.apiary.io/
    addon_api_url = 'https://addons-ecs.forgesvc.net/api/v2/addon'
    slugs_url = ('https://raw.githubusercontent.com/layday/instascrape/data/'
                 'curseforge-slugs-v2.compact.json')   # v2

    _slugs: Optional[Dict[str, str]] = None

    @classmethod
    def decompose_url(cls, value: str) -> Optional[Tuple[str, str]]:
        url = URL(value)
        if (url.host == 'www.wowace.com'
                and len(url.parts) > 2
                and url.parts[1] == 'projects'):
            return (cls.source, url.parts[2].lower())
        elif (url.host == 'www.curseforge.com'
                and len(url.parts) > 3
                and url.parts[1:3] == ('wow', 'addons')):
            return (cls.source, url.parts[3].lower())

    async def synchronise(self) -> CurseResolver:
        if self._slugs is None:
            self._slugs = await self._cache_json_response(self.slugs_url, 8, 'hours')
        return self

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        slugs = cast(dict, self._slugs)
        ids_from_defns = {d: slugs.get(d.name, d.name) for d in defns}
        numeric_ids = list({i for i in ids_from_defns.values() if i.isdigit()})
        async with self.web_client.post(self.addon_api_url, json=numeric_ids) as response:
            if response.status == 404:
                json_response = []       # type: ignore
            else:
                json_response = await response.json()

        api_results = {str(r['id']): r for r in json_response}
        coros = (self.resolve_one(d, api_results.get(i))
                 for d, i in ids_from_defns.items())
        results = dict(zip(defns, await gather(coros)))
        return results

    @validate_strategy
    async def resolve_one(self, defn: Defn, metadata: Optional[dict]) -> m.Pkg:
        if not metadata:
            raise E.PkgNonexistent
        elif not metadata['latestFiles']:
            raise E.PkgFileUnavailable('no files available for download')

        def is_not_libless(f: dict) -> bool:
            # There's also an 'isAlternate' field that's missing some
            # 50 lib-less files from c. 2008.  'exposeAsAlternative' is
            # absent from the /file endpoint
            return not f['exposeAsAlternative']

        classic_version_prefix = '1.13'
        flavor = 'wow_classic' if self.config.is_classic else 'wow_retail'

        def is_compatible_with_game_version(f: dict) -> bool:
            # Files can belong both to retail and classic
            # but ``gameVersionFlavor`` can only be one of
            # 'wow_retail' or 'wow_classic'.  To spice things up,
            # ``gameVersion`` might not be populated so we still have to
            # check the value of ``gameVersionFlavor``
            return (f['gameVersionFlavor'] == flavor
                    or any(v.startswith(classic_version_prefix) is self.config.is_classic
                           for v in f['gameVersion']))

        # 1 = stable; 2 = beta; 3 = alpha
        if defn.strategy is Strategies.latest:
            def has_release_type(f: dict) -> bool: return True
        elif defn.strategy is Strategies.curse_latest_beta:
            def has_release_type(f: dict) -> bool: return f['releaseType'] == 2
        elif defn.strategy is Strategies.curse_latest_alpha:
            def has_release_type(f: dict) -> bool: return f['releaseType'] == 3
        else:
            def has_release_type(f: dict) -> bool: return f['releaseType'] == 1

        files = (f for f in metadata['latestFiles']
                 if is_not_libless(f)
                 and is_compatible_with_game_version(f)
                 and has_release_type(f))
        try:
            # The ``id`` is just a counter so we don't have to go digging around dates
            file = max(files, key=itemgetter('id'))
        except ValueError:
            raise E.PkgFileUnavailable(f'no files compatible with {self.config.game_flavour} '
                                       f'using {defn.strategy.name!r} strategy')

        # 1 = embedded library
        # 2 = optional dependency
        # 3 = required dependency
        # 4 = tool
        # 5 = incompatible
        # 6 = include (wat)
        deps = [m.PkgDep(id=d['addonId']) for d in file['dependencies']
                if d['type'] == 3]

        return m.Pkg(origin=self.source,
                     id=metadata['id'],
                     slug=metadata['slug'],
                     name=metadata['name'],
                     description=metadata['summary'],
                     url=metadata['websiteUrl'],
                     file_id=file['id'],
                     download_url=file['downloadUrl'],
                     date_published=file['fileDate'],
                     version=file['displayName'],
                     options=m.PkgOptions(strategy=defn.strategy.name),
                     deps=deps)


class WowiResolver(Resolver, _FileCacheMixin):
    source = 'wowi'
    name = 'WoWInterface'
    strategies = {Strategies.default}

    # Reference: https://api.mmoui.com/v3/globalconfig.json
    # There's also a v4 API for the as yet unreleased Minion v4 which I
    # would assume is unstable.  From a cursory glance it looks like they've
    # renamed all fields in camelCase; change the type of
    # some numeric-only fields to a number; and removed 'UISiblings'
    # which was always empty
    list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    _files: Optional[Dict[str, dict]] = None

    @classmethod
    def decompose_url(cls, value: str) -> Optional[Tuple[str, str]]:
        url = URL(value)
        if (url.host in {'wowinterface.com', 'www.wowinterface.com'}
                and len(url.parts) == 3
                and url.parts[1] == 'downloads'):
            if url.name == 'landing.php':
                id_ = url.query.get('fileid')
                if id_:
                    return (cls.source, id_)
            elif url.name == 'fileinfo.php':
                id_ = url.query.get('id')
                if id_:
                    return (cls.source, id_)
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)(?:-(?P<name>[^\.]+))?',
                                 url.name)
                if match:
                    id_, slug = match.groups()
                    if slug:
                        id_ = slugify(f'{id_} {slug}')
                    return (cls.source, id_)

    async def synchronise(self) -> WowiResolver:
        if self._files is None:
            files = await self._cache_json_response(self.list_api_url, 3600)
            self._files = {i['UID']: i for i in files}
        return self

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        files = cast(dict, self._files)
        ids_from_defns = {d: ''.join(takewhile(str.isdigit, d.name)) for d in defns}
        numeric_ids = {i for i in ids_from_defns.values() if i.isdigit()}
        url = self.details_api_url / f'{",".join(numeric_ids)}.json'
        async with self.web_client.get(url) as response:
            if response.status == 404:
                json_response = []       # type: ignore
            else:
                json_response = await response.json()

        api_results = {r['UID']: {**files[r['UID']], **r} for r in json_response}
        coros = (self.resolve_one(d, api_results.get(i))
                 for d, i in ids_from_defns.items())
        results = dict(zip(defns, await gather(coros)))
        return results

    @validate_strategy
    async def resolve_one(self, defn: Defn, metadata: Optional[dict]) -> m.Pkg:
        if not metadata:
            raise E.PkgNonexistent
        if metadata['UIPending'] == '1':
            raise E.PkgFileUnavailable('new file awaiting approval')

        # WoWInterface has recently moved to a multi-file setup like CurseForge
        # and they botched it, like, completely.  Classic add-on files aren't
        # included in the API response and 'UICompatibility' can include either
        # retail or classic or both (!) but the file linked to from the API
        # is always the retail version.  You can have an add-on that is
        # nominally compatible with classic but if you were to install it
        # for classic you'd get the retail version and if you were to attempt
        # to install it for retail instawow might report that it's
        # incompatible.  The way they rushed this out without an ounce of
        # thought or care for their own add-on manager and considering it
        # hasn't received a proper update closing in on 3 years makes me
        # doubt that a fix is on the horizon.  There isn't a sensible way for
        # instawow to discern what version of the game a file's compatible with
        # so it opts to not opt to do anything at all.  Good luck and godspeed.

        return m.Pkg(origin=self.source,
                     id=metadata['UID'],
                     slug=slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
                     name=metadata['UIName'],
                     description=metadata['UIDescription'],
                     url=metadata['UIFileInfoURL'],
                     file_id=metadata['UIMD5'],
                     download_url=metadata['UIDownload'],
                     date_published=metadata['UIDate'],
                     version=metadata['UIVersion'],
                     options=m.PkgOptions(strategy=defn.strategy.name))


class TukuiResolver(Resolver):
    source = 'tukui'
    name = 'Tukui'
    strategies = {Strategies.default}

    api_url = URL('https://www.tukui.org/api.php')

    retail_uis = {'-1': 'tukui', '-2': 'elvui',
                  'tukui': 'tukui', 'elvui': 'elvui'}

    @classmethod
    def decompose_url(cls, value: str) -> Optional[Tuple[str, str]]:
        url = URL(value)
        if (url.host == 'www.tukui.org'
                and url.path in {'/addons.php', '/classic-addons.php', '/download.php'}):
            id_or_slug = url.query.get('id') or url.query.get('ui')
            if id_or_slug:
                return (cls.source, id_or_slug)

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        results = await gather(self.resolve_one(d) for d in defns)
        return dict(zip(defns, results))

    @validate_strategy
    async def resolve_one(self, defn: Defn) -> m.Pkg:
        name = ui_name = self.retail_uis.get(defn.name)
        if not name:
            name = ''.join(takewhile('-'.__ne__, defn.name))

        if self.config.is_classic:
            query = 'classic-addon'
        elif ui_name:
            query = 'ui'
        else:
            query = 'addon'

        url = self.api_url.with_query({query: name})
        async with self.web_client.get(url) as response:
            if not response.content_length:
                raise E.PkgNonexistent
            addon = await response.json(content_type='text/html')

        return m.Pkg(origin=self.source,
                     id=addon['id'],
                     slug=ui_name or slugify(f'{addon["id"]} {addon["name"]}'),
                     name=addon['name'],
                     description=addon['small_desc'],
                     url=addon['web_url'],
                     file_id=addon['version'],
                     download_url=addon['url'],
                     date_published=datetime.fromisoformat(addon['lastupdate']),
                     version=addon['version'],
                     options=m.PkgOptions(strategy=defn.strategy.name))


class InstawowResolver(Resolver):
    source = 'instawow'
    name = 'instawow'
    strategies = {Strategies.default}

    _addons = {('0', 'weakauras-companion'),
               ('1', 'weakauras-companion-autoupdate')}

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        results = await gather(self.resolve_one(d) for d in defns)
        return dict(zip(defns, results))

    @validate_strategy
    async def resolve_one(self, defn: Defn) -> m.Pkg:
        try:
            id_, slug = next(p for p in self._addons if defn.name in p)
        except StopIteration:
            raise E.PkgNonexistent

        from .wa_updater import WaCompanionBuilder

        builder = WaCompanionBuilder(self.manager)
        if id_ == '1':
            try:
                await builder.build()
            except ValueError as error:
                raise E.PkgFileUnavailable('account name not provided') from error

        return m.Pkg(origin=self.source,
                     id=id_,
                     slug=slug,
                     name='WeakAuras Companion',
                     description='A WeakAuras Companion clone.',
                     url='https://github.com/layday/instawow',
                     file_id=await t(builder.checksum)(),
                     download_url=builder.file_out.as_uri(),
                     date_published=datetime.now(),
                     version='1.0.0',
                     options=m.PkgOptions(strategy=defn.strategy.name))



from __future__ import annotations

import asyncio
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional, Sequence

from loguru import logger
from pydantic import BaseModel, BaseSettings, Extra, validator
from yarl import URL

from .config import BaseConfig
from .utils import ManagerAttrAccessMixin, bucketise, run_in_thread as t

if TYPE_CHECKING:
    from .manager import Manager


metadata_api = URL('https://data.wago.io/api/check/weakauras')
raw_api = URL('https://data.wago.io/api/raw/encoded')


class BuilderConfig(BaseConfig):
    account: str

    class Config:
        env_prefix = 'WAC_'


class AuraEntry(BaseModel):
    id: str
    uid: str
    parent: Optional[str]
    url: URL
    version: int
    semver: Optional[str]
    ignore_wago_update: bool = False

    class Config:
        arbitrary_types_allowed = True
        fields = {'ignore_wago_update': {'alias': 'ignoreWagoUpdate'}}

    @validator('url', pre=True)
    def __convert_url(cls, value: str) -> URL:
        return URL(value)

    @classmethod
    def from_lua_ast(cls, values: dict) -> Optional[AuraEntry]:
        return cls(**values) if values.get('url') else None


class ApiMetadata__Changelog(BaseModel):
    format: Optional[str]
    text: str = ''

    class Config:
        extra = Extra.forbid


class ApiMetadata(BaseModel):
    id: str
    name: str
    slug: str
    url: str
    created: str
    modified: str
    game: str
    fork_of: Optional[str]
    username: str
    version: int
    version_string: str
    changelog: ApiMetadata__Changelog
    region_type: Optional[str]

    class Config:
        extra = Extra.forbid
        fields = {'id': {'alias': '_id'},
                  'fork_of': {'alias': 'forkOf'},
                  'version_string': {'alias': 'versionString'},
                  'region_type': {'alias': 'regionType'}}


class _OutdatedAura(NamedTuple):
    slug: str
    existing_auras: List[AuraEntry]
    metadata: ApiMetadata
    import_string: str


class WaCompanionBuilder(ManagerAttrAccessMixin):
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager: Manager) -> None:
        self.manager = manager
        self.builder_dir = self.manager.config.plugin_dir / __name__
        self.file_out = self.builder_dir / 'WeakAurasCompanion.zip'

    @staticmethod
    def extract_auras(source: str) -> Dict[str, List[AuraEntry]]:
        from lupa import LuaRuntime

        lua_runtime = LuaRuntime()
        table = lua_runtime.eval(source[source.find('= ') + 1 :])
        raw_auras = map(dict, table['displays'].values())
        auras = filter(None, map(AuraEntry.from_lua_ast, raw_auras))
        aura_groups = {k: v
                       for k, v in bucketise(auras, key=lambda a: a.url.parts[1]).items()
                       if not any(i.ignore_wago_update for i in v)}
        return aura_groups

    def extract_installed_auras(self, account: str) -> Dict[str, List[AuraEntry]]:
        import time

        accounts = self.config.addon_dir.parents[1] / 'WTF/Account'
        saved_vars = accounts / account / 'SavedVariables/WeakAuras.lua'

        start = time.perf_counter()
        aura_groups = self.extract_auras(saved_vars.read_text(encoding='utf-8'))
        logger.debug(f'auras extracted in {time.perf_counter() - start}s')
        return aura_groups

    async def get_wago_aura_metadata(self, aura_ids: Sequence[str]) -> List[ApiMetadata]:
        url = metadata_api.with_query({'ids': ','.join(aura_ids)})
        async with self.web_client.get(url) as response:
            metadata = await response.json()

        results = dict.fromkeys(aura_ids)
        for item in (ApiMetadata(**i) for i in metadata):
            if item.slug in results:
                results[item.slug] = item
            else:
                logger.info(f'extraneous {item.slug!r} slug in metadata')
        return list(results.values())

    async def get_wago_aura_import_string(self, aura_id: str) -> str:
        url = raw_api.with_query({'id': aura_id})
        async with self.web_client.get(url) as response:
            return await response.text()

    async def get_outdated(self, aura_groups: Dict[str, List[AuraEntry]]) -> List[_OutdatedAura]:
        if not aura_groups:
            return []

        aura_metadata = await self.get_wago_aura_metadata(list(aura_groups))
        auras_with_metadata = filter(lambda v: v[1], zip(aura_groups.items(), aura_metadata))
        outdated_auras = [((s, w), r)
                          for (s, w), r in auras_with_metadata
                          if r.version > next(chain((a for a in w if not a.parent), w)).version]
        new_auras = await asyncio.gather(*(self.get_wago_aura_import_string(r.id)
                                           for _, r in outdated_auras))
        return [_OutdatedAura(s, w, r, p)
                for ((s, w), r), p in zip(outdated_auras, new_auras)]

    def make_addon(self, outdated: List[_OutdatedAura]) -> None:
        from jinja2 import Environment, FileSystemLoader
        from zipfile import ZipFile, ZipInfo

        tpl_dir = Path(__file__).parent / 'wa_templates'
        jinja_env = Environment(loader=FileSystemLoader(str(tpl_dir)),
                                trim_blocks=True, lstrip_blocks=True)

        with ZipFile(self.file_out, 'w') as addon_zip:
            def write_tpl(filename: str, ctx: dict) -> None:
                # Not using a plain string as the first argument to ``writestr``
                # 'cause the timestamp is set dynamically by default, which
                # renders the build rather - shall we say - unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                tpl = jinja_env.get_template(filename)
                addon_zip.writestr(zip_info, tpl.render(**ctx))

            write_tpl('data.lua',
                      {'was': [(o.metadata.slug,
                                {'name': o.metadata.name,
                                 'author': o.metadata.username,
                                 'encoded': o.import_string,
                                 'wagoVersion': o.metadata.version,
                                 'wagoSemver': o.metadata.version_string,
                                 'versionNote': o.metadata.changelog.text})
                               for o in outdated],
                       'uids': [(a.uid, a.url.parts[1])
                                for o in outdated
                                for a in o.existing_auras],
                       'ids':  [(a.id, a.url.parts[1])
                                for o in outdated
                                for a in o.existing_auras],
                       'stash': []})    # Send to WAC not supported - always empty
            write_tpl('init.lua', {})
            write_tpl('WeakAurasCompanion.toc',
                      {'interface': '11303' if self.config.is_classic else '80205'})

    async def build(self, account: Optional[str] = None) -> None:
        config = BuilderConfig(account=account)
        auras = await t(self.extract_installed_auras)(config.account)
        outdated_auras = await self.get_outdated(auras)

        await t(self.builder_dir.mkdir)(exist_ok=True)
        await t(self.make_addon)(outdated_auras)

    def checksum(self) -> str:
        from hashlib import sha256
        return sha256(self.file_out.read_bytes()).hexdigest()



from __future__ import annotations

import asyncio
from collections import defaultdict, deque
from datetime import datetime, timedelta
from functools import reduce
from pathlib import Path
import re
from typing import (TYPE_CHECKING, AbstractSet, Any, Awaitable, Callable, Iterable, List,
                    NamedTuple, Optional, Sequence, Tuple, Type, TypeVar, Union)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal       # type: ignore

if TYPE_CHECKING:
    import prompt_toolkit.shortcuts.progress_bar.base as pbb
    from .manager import CliManager


class ManagerAttrAccessMixin:

    def __getattr__(self, name: str) -> Any:
        return getattr(self.manager, name)


class _TocEntry(NamedTuple):

    key: str
    value: str

    def __bool__(self) -> bool:
        return bool(self.value)


class TocReader:
    """Extracts key–value pairs from TOC files."""

    def __init__(self, contents: str) -> None:
        possible_entries = (map(str.strip, e.lstrip('#').partition(':')[::2])
                            for e in contents.splitlines()
                            if e.startswith('##'))
        self.entries = {k: v for k, v in possible_entries if k}
        self.default = ''

    def __getitem__(self, key: Union[str, Tuple[str, ...]]) -> _TocEntry:
        if isinstance(key, tuple):
            try:
                return next(filter(None, (self[k] for k in key)))
            except StopIteration:
                key = key[0]
        return _TocEntry(key, self.entries.get(key, self.default))

    @classmethod
    def from_path(cls, path: Path) -> TocReader:
        return cls(path.read_text(encoding='utf-8-sig', errors='replace'))

    @classmethod
    def from_path_name(cls, path: Path) -> TocReader:
        return cls.from_path(path / f'{path.name}.toc')


O = TypeVar('O')


class cached_property:

    def __init__(self, f: Callable) -> None:
        self.f = f

    def __get__(self, o: Optional[O], t: Type[O]) -> Any:
        if o is None:
            return self.f
        else:
            o.__dict__[self.f.__name__] = v = self.f(o)
            return v


def bucketise(iterable: Iterable, key: Callable = (lambda v: v)) -> defaultdict:
    "Place the elements of ``iterable`` into a bucket according to ``key``."
    bucket = defaultdict(list)      # type: ignore
    for value in iterable:
        bucket[key(value)].append(value)
    return bucket


def dict_merge(*args: dict) -> dict:
    "Right merge any number of ``dict``s."
    return reduce(lambda p, n: {**p, **n}, args, {})


def merge_intersecting_sets(it: Iterable[AbstractSet]) -> Iterable[AbstractSet]:
    "Recursively merge intersecting sets in a collection."
    many_sets = deque(it)
    while True:
        try:
            this_set = many_sets.popleft()
        except IndexError:
            return
        while True:
            for other_set in many_sets:
                if this_set & other_set:
                    # The in-place operator will mutate unfrozen sets in the original collection
                    this_set = this_set | other_set
                    many_sets.remove(other_set)
                    break
            else:
                break
        yield this_set


async def gather(it: Iterable, return_exceptions: bool = True) -> List[Any]:
    return await asyncio.gather(*it, return_exceptions=return_exceptions)


def run_in_thread(fn: Callable) -> Callable[..., Awaitable]:
    return lambda *a, **k: asyncio.get_running_loop().run_in_executor(None, lambda: fn(*a, **k))


_match_loweralphanum = re.compile(r'[^0-9a-z ]')


def slugify(text: str) -> str:
    "Convert an add-on name into a lower-alphanumeric slug."
    return '-'.join(_match_loweralphanum.sub(' ', text.casefold()).split())


def tabulate(rows: Sequence, *, max_col_width: int = 60) -> str:
    "Produce an ASCII table from equal-length elements in a sequence."
    from textwrap import fill

    def apply_max_col_width(value):
        return fill(str(value), width=max_col_width, max_lines=1)

    def calc_resultant_col_widths(rows):
        cols = zip(*rows)
        return [max(map(len, c)) for c in cols]

    rows = [tuple(map(apply_max_col_width, r)) for r in rows]
    head, *tail = rows

    base_template = ' '.join(f'{{{{{{0}}{w}}}}}'
                             for w in calc_resultant_col_widths(rows))
    row_template = base_template.format(':<')
    table = '\n'.join((base_template.format(':^').format(*head),
                       base_template.format(f'0:-<').format(''),
                       *(row_template.format(*r) for r in tail),))
    return table


def make_progress_bar(**kwargs: Any) -> pbb.ProgressBar:
    "A ``ProgressBar`` with download progress expressed in megabytes."
    from contextlib import contextmanager
    from prompt_toolkit.formatted_text import HTML
    import prompt_toolkit.shortcuts.progress_bar.base as pbb
    from prompt_toolkit.shortcuts.progress_bar import formatters

    # There is a race condition in the bar's shutdown logic
    # where the bar-drawing thread is left to run indefinitely
    # if the prompt_toolkit app is not (yet) running
    # but is scheduled to run when ``__exit__``ing the bar.
    # If an exception occurs early in the execution cycle or if execution
    # finishes before prompt_toolkit is able to crank the app,
    # instawow will hang.  This is my ham-fisted attempt to work
    # around all that by relocating the ``.exit()`` call.  The progress bar
    # 'signals' to the thread that the app is ready to exit
    # and the app is exited from inside the thread.
    # To avoid having to monkey-patch ``run`` which is nested
    # inside ``__enter__``, we overwrite ``pbb._auto_refresh_context``.

    class ProgressBar(pbb.ProgressBar):
        def __exit__(self, *args):
            if self._has_sigwinch:
                self._loop.add_signal_handler(pbb.signal.SIGWINCH, self._previous_winch_handler)

            # Signal to ``_auto_refresh_context`` that it should exit the app
            self.app._should_exit = True
            self._thread.join()
            del self.app._should_exit

    _bar_refresh_interval = .1

    @contextmanager
    def _auto_refresh_context(app, _refresh_interval=None):
        done = [False]

        def run():
            while not done[0]:
                pbb.time.sleep(_bar_refresh_interval)
                app.invalidate()
                if getattr(app, '_should_exit', False) and app.is_running:
                    app.exit()

        t = pbb.threading.Thread(target=run)
        t.daemon = True
        t.start()

        try:
            yield
        finally:
            done[0] = True

    pbb._auto_refresh_context = _auto_refresh_context

    class ProgressInMB(formatters.Progress):
        template = f'{formatters.Progress.template}MB'

        def format(self, _bar, counter, _width):
            html = HTML(self.template)
            return html.format(**{f: f'{getattr(counter, f) / 2 ** 20:.1f}'
                                  for f in ('current', 'total')})

    formatters = [formatters.Label(),
                  formatters.Text(' '),
                  formatters.Percentage(),
                  formatters.Text(' '),
                  formatters.Bar(),
                  formatters.Text(' '),
                  ProgressInMB(),
                  formatters.Text(' '),
                  formatters.Text('eta [', style='class:time-left'),
                  formatters.TimeLeft(),
                  formatters.Text(']', style='class:time-left'),
                  formatters.Text(' '),]
    progress_bar = ProgressBar(formatters=formatters, **kwargs)
    return progress_bar


def shasum(*values: str) -> str:
    "Base-16-encode a string using SHA-256 truncated to 32 characters."
    from hashlib import sha256

    return sha256(''.join(values).encode()).hexdigest()[:32]


def is_not_stale(path: Path, ttl: int, unit: str = 'seconds') -> bool:
    "Check if a file is older than ``ttl``."
    mtime = path.exists() and path.stat().st_mtime
    return mtime > 0 and (datetime.now() - datetime.fromtimestamp(mtime)) < timedelta(**{unit: ttl})


def get_version() -> str:
    try:
        import importlib.metadata as importlib_metadata
    except ImportError:
        import importlib_metadata       # type: ignore

    try:
        return importlib_metadata.version(__package__)
    except importlib_metadata.PackageNotFoundError:
        return 'dev'


def is_outdated(manager: CliManager) -> bool:
    """Check against PyPI to see if `instawow` is outdated.

    The response is cached for 24 hours.
    """
    __version__ = get_version()
    if __version__ == 'dev':
        return False

    def parse_version(version: str) -> Tuple[int, ...]:
        return tuple(map(int, version.split('.')[:3]))

    cache_file = manager.config.config_dir / '.pypi_version'
    if is_not_stale(cache_file, 1, 'days'):
        version = cache_file.read_text(encoding='utf-8')
    else:
        from aiohttp.client import ClientError

        async def get_metadata() -> dict:
            url = 'https://pypi.org/pypi/instawow/json'
            async with manager.web_client.get(url) as response:
                return await response.json()

        try:
            version = manager.run(get_metadata())['info']['version']
        except ClientError:
            version = __version__
        else:
            cache_file.write_text(version, encoding='utf-8')

    return parse_version(version) > parse_version(__version__)


def setup_logging(logger_dir: Path, level: Union[int, str] = 'INFO') -> int:
    from loguru import logger

    handler = {'sink': logger_dir / 'error.log',
               'level': level,
               'rotation': '1 MB',
               'enqueue': True}
    handler_id, = logger.configure(handlers=(handler,))
    return handler_id



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, List

import pydantic
from pydantic import create_model
import sqlalchemy.exc
from sqlalchemy.ext.declarative import DeclarativeMeta, as_declarative
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKeyConstraint, UniqueConstraint
from sqlalchemy.types import DateTime, Integer, String

if TYPE_CHECKING:
    import sqlalchemy.base

    from .resolvers import Defn


class _BaseCoercer(pydantic.BaseModel):
    """The coercer is used inside the declarative constructor to type-cast
    values _prior to_ inserts.  SQLAlchemy delegates this function to
    the DB API -- see https://stackoverflow.com/a/8980982.  We need this to
    happen in advance to be able to compare values with their in-database
    counterparts.  It also saves us the trouble of having to manually
    parse dates and the like.
    """

    class Config:
        extra = pydantic.Extra.allow
        max_anystr_length = 2 ** 32
        orm_mode = True


class _BaseTableMeta(DeclarativeMeta):
    def __init__(cls, *args: Any) -> None:
        super().__init__(*args)
        try:
            columns = inspect(cls).columns
        except sqlalchemy.exc.NoInspectionAvailable:
            pass
        else:
            cls.Coercer = create_model(
                f'{cls.__name__}Coercer',
                __base__=_BaseCoercer,
                **{c.name: (c.type.python_type, ...) for c in columns    # type: ignore
                   if not c.foreign_keys and not c.name.startswith('_')})


@as_declarative(constructor=None, metaclass=_BaseTableMeta)
class _BaseTable:
    Coercer: ClassVar

    def __init__(self, **kwargs: Any) -> None:
        intermediate_obj = self.Coercer.parse_obj(kwargs)
        for k, v in intermediate_obj:
            setattr(self, k, v)

ModelBase = _BaseTable


class Pkg(_BaseTable):
    __tablename__ = 'pkg'

    origin = Column(String, primary_key=True)
    id = Column(String, primary_key=True)
    slug = Column(String, nullable=False)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)
    url = Column(String, nullable=False)
    file_id = Column(String, nullable=False)
    download_url = Column(String, nullable=False)
    date_published = Column(DateTime, nullable=False)
    version = Column(String, nullable=False)
    folders = relationship('PkgFolder', cascade='all, delete-orphan', backref='pkg')
    options = relationship('PkgOptions', cascade='all, delete-orphan', uselist=False)
    deps = relationship('PkgDep', cascade='all, delete-orphan', backref='pkg')

    def to_defn(self) -> Defn:
        from .resolvers import Defn, Strategies

        return Defn(self.origin, self.slug, Strategies[self.options.strategy])


class PkgFolder(_BaseTable):
    __tablename__ = 'pkg_folder'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),)

    name = Column(String, primary_key=True)
    pkg_origin = Column(String, nullable=False)
    pkg_id = Column(String, nullable=False)


class PkgOptions(_BaseTable):
    __tablename__ = 'pkg_options'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),)

    strategy = Column(String, nullable=False)
    pkg_origin = Column(String, primary_key=True)
    pkg_id = Column(String, primary_key=True)


class PkgDep(_BaseTable):
    __tablename__ = 'pkg_dep'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),
                      UniqueConstraint('id', 'pkg_origin', 'pkg_id',
                                       name='uq_id_per_foreign_key_constr'))

    _id = Column(Integer, primary_key=True)
    id = Column(String, nullable=False)
    pkg_origin = Column(String, nullable=False)
    pkg_id = Column(String, nullable=False)


class _PkgModel(Pkg.Coercer):
    folders: List[PkgFolder.Coercer]
    options: PkgOptions.Coercer
    deps: List[PkgDep.Coercer]


class MultiPkgModel(pydantic.BaseModel):
    __root__: List[_PkgModel]

    @classmethod
    def from_orm(cls, values: List[Pkg]) -> MultiPkgModel:
        return cls.parse_obj([_PkgModel.from_orm(v) for v in values])


def should_migrate(engine: sqlalchemy.base.Engine, version: str) -> bool:
    """Check if the database version is the same as `version`;
    if not a migration would be required.
    """
    with engine.begin() as conn:
        try:
            current = conn.execute('SELECT version_num '
                                   'FROM alembic_version '
                                   'WHERE version_num = (?)', version).scalar()
        except sqlalchemy.exc.OperationalError:
            return True
        else:
            if not current:
                return True
    return False


def is_pkg(value: Any) -> bool:
    return isinstance(value, Pkg)



from __future__ import annotations

from functools import partial, wraps
from itertools import chain, islice
from pathlib import Path
from typing import (TYPE_CHECKING, Callable, Dict, Iterable, List, Optional, Sequence, Tuple,
                    Union, cast)

import click

from . import exceptions as E, models
from .resolvers import Defn, Strategies
from .utils import (TocReader, bucketise, cached_property, get_version, is_outdated, setup_logging,
                    tabulate)

if TYPE_CHECKING:
    from .manager import CliManager


class Report:
    _success = click.style('✓', fg='green')
    _failure = click.style('✗', fg='red')
    _warning = click.style('!', fg='blue')

    def __init__(self, results: Dict[Defn, E.ManagerResult],
                 filter_fn: Callable = (lambda _: True),
                 report_outdated: bool = True) -> None:
        self.results = results
        self.filter_fn = filter_fn
        self.report_outdated = report_outdated

    @property
    def code(self) -> int:
        return any(r for r in self.results.values()
                   if (isinstance(r, (E.ManagerError, E.InternalError))
                       and self.filter_fn(r)))

    def __str__(self) -> str:
        def _adorn_result(result: E.ManagerResult) -> str:
            if isinstance(result, E.InternalError):
                return self._warning
            elif isinstance(result, E.ManagerError):
                return self._failure
            return self._success

        return '\n'.join((f'{_adorn_result(r)} {click.style(str(a), bold=True)}\n'
                          f'  {r.message}')
                         for a, r in self.results.items()
                         if self.filter_fn(r))

    def generate(self) -> None:
        if self.report_outdated:
            manager = click.get_current_context().obj.m
            if is_outdated(manager):
                click.echo(f'{self._warning} instawow is out of date')

        report = str(self)
        if report:
            click.echo(report)

    def generate_and_exit(self) -> None:
        self.generate()
        ctx = click.get_current_context()
        ctx.exit(self.code)


def parse_into_defn(manager, value: Union[str, Sequence[str]], *,
                    raise_invalid: bool = True) -> Union[Defn, List[Defn]]:
    if isinstance(value, (list, tuple)):
        # Bucketise to remove dupes
        return list(bucketise(parse_into_defn(manager, v, raise_invalid=raise_invalid)
                              for v in value))

    value = cast(str, value)
    if ':' not in value:
        if raise_invalid:
            raise click.BadParameter(value)

        parts = ('*', value)
    else:
        for resolver in manager.resolvers.values():
            parts = resolver.decompose_url(value)
            if parts:
                break
        else:
            parts = value.partition(':')[::2]
    return Defn(*parts)


def parse_into_defn_with_strategy(manager, value: List[Tuple[str, str]]) -> List[Defn]:
    defns = parse_into_defn(manager, [d for _, d in value])
    strategies = (Strategies[s] for s, _ in value)
    return list(map(Defn.with_strategy, defns, strategies))


def export_to_csv(pkgs: List[models.Pkg]) -> str:
    "Export packages to CSV."
    import csv
    import io

    buffer = io.StringIO()
    writer = csv.writer(buffer)
    writer.writerow(('defn', 'strategy'))
    writer.writerows((p.to_defn(), p.options.strategy) for p in pkgs)
    return buffer.getvalue()


def import_from_csv(manager, value: Iterable[str]) -> List[Defn]:
    "Import definitions from CSV."
    import csv

    rows = [(b, a) for a, b in islice(csv.reader(value), 1, None)]
    return parse_into_defn_with_strategy(manager, rows)


def _pass_manager(fn: Callable) -> Callable:
    @wraps(fn)
    def wrapper(*args, **kwargs):
        ctx = click.get_current_context()
        return fn(ctx.obj.m, *args, **kwargs)

    return wrapper


def _callbackify(fn: Callable) -> Callable:
    return lambda c, p, v: fn(c.obj.m, v)


def _combine_addons(fn: Callable) -> Callable:
    def combine(ctx, param, value):
        addons = ctx.params.setdefault('addons', [])
        if value:
            addons.extend(fn(ctx.obj.m, value))

    return combine


def _show_version(ctx: click.Context, param, value: bool) -> None:
    if value:
        __version__ = get_version()
        click.echo(f'instawow, version {__version__}')
        ctx.exit()


class _InstallCommand(click.Command):
    def format_epilog(self, ctx, formatter):
        with formatter.section('Strategies'):
            formatter.write_dl((s.name, s.value) for s in islice(Strategies, 1, None))


@click.group(context_settings={'help_option_names': ('-h', '--help')})
@click.option('--version',
              is_flag=True, default=False,
              expose_value=False, is_eager=True,
              callback=_show_version,
              help='Show the version and exit.')
@click.option('--debug',
              is_flag=True, default=False,
              help='Log more things.')
@click.pass_context
def main(ctx, debug: bool) -> None:
    "Add-on manager for World of Warcraft."
    if not ctx.obj:
        @object.__new__
        class ManagerSingleton:
            @cached_property
            def m(self) -> CliManager:
                from .config import Config
                from .manager import CliManager, prepare_db_session

                while True:
                    try:
                        config = Config.read().ensure_dirs()
                    except FileNotFoundError:
                        ctx.invoke(write_config)
                    else:
                        break

                setup_logging(config.logger_dir, 'DEBUG' if debug else 'INFO')
                db_session = prepare_db_session(config)
                manager = CliManager(config, db_session)
                return manager

        ctx.obj = ManagerSingleton


@main.command(cls=_InstallCommand)
@click.option('--replace', '-o',
              is_flag=True, default=False,
              help='Replace existing add-ons.')
@click.option('--import', '-i',
              type=click.File(encoding='utf-8'),
              callback=_combine_addons(import_from_csv),
              expose_value=False,
              help='Install add-ons from CSV.')
@click.option('--with-strategy', '-s',
              multiple=True,
              type=(click.Choice([s.name for s in Strategies]), str),
              callback=_combine_addons(parse_into_defn_with_strategy),
              expose_value=False,
              metavar='<STRATEGY ADDON>...',
              help='A strategy followed by an add-on definition.  '
                   'Use this if you want to install an add-on with a '
                   'strategy other than the default one.  '
                   'Repeatable.')
@click.argument('addons',
                nargs=-1, callback=_combine_addons(parse_into_defn),
                expose_value=False)
@_pass_manager
def install(manager, addons: Sequence[Defn], replace: bool) -> None:
    "Install add-ons."
    if not addons:
        raise click.UsageError('Require at least one of "ADDONS", '
                               '"--with-strategy" or "--import".')

    deduped_defns = [v for v, *_ in
                     bucketise(addons, lambda d: (d.source, d.name)).values()]
    results = manager.install(deduped_defns, replace)
    Report(results).generate_and_exit()


@main.command()
@click.argument('addons',
                nargs=-1, callback=_callbackify(parse_into_defn))
@_pass_manager
def update(manager, addons: Sequence[Defn]) -> None:
    "Update installed add-ons."
    def report_filter(result):
        # Hide package from output if up to date and ``update`` was invoked without args
        return addons or not isinstance(result, E.PkgUpToDate)

    defns = addons
    if not defns:
        defns = [p.to_defn() for p in manager.db_session.query(models.Pkg).all()]

    results = manager.update(defns)
    Report(results, report_filter).generate_and_exit()


@main.command()
@click.argument('addons',
                nargs=-1, required=True, callback=_callbackify(parse_into_defn))
@_pass_manager
def remove(manager, addons: Sequence[Defn]) -> None:
    "Uninstall add-ons."
    results = manager.remove(addons)
    Report(results).generate_and_exit()


@main.command()
@click.option('--auto', '-a',
              is_flag=True, default=False,
              help='Do not ask for user confirmation.')
@click.pass_context
def reconcile(ctx, auto: bool) -> None:
    "Reconcile add-ons."
    from .matchers import match_toc_ids, match_dir_names, get_folders
    from .models import Pkg, is_pkg
    from .prompts import PkgChoice, confirm, select, skip

    preamble = '''\
Use the arrow keys to navigate, <o> to open an add-on in your browser,
enter to make a selection and <s> to skip to the next item.

Versions that differ from the installed version or differ between choices
are highlighted in purple.

The reconciler will do a first pass of all of your add-ons looking for
source IDs in TOC files.  If it is unable to reconcile all of your add-ons
it will perform a second pass to match add-on folders against the CurseForge
and WoWInterface catalogues.

Selected add-ons _will_ be reinstalled.
'''

    manager = ctx.obj.m

    def prompt_one(addons, pkgs):
        def create_choice(pkg):
            defn = pkg.to_defn()
            title = [('', str(defn)),
                     ('', '=='),
                     ('class:highlight-sub' if highlight_version else '', pkg.version),]
            return PkgChoice(title, defn, pkg=pkg)

        # Highlight version if there's multiple of them
        highlight_version = len({i.version for i in chain(addons, pkgs)}) > 1
        choices = list(chain(map(create_choice, pkgs), (skip,)))
        addon = addons[0]
        # Using 'unsafe_ask' to let ^C bubble up
        selection = select(f'{addon.name} [{addon.version or "?"}]', choices).unsafe_ask()
        return selection

    def prompt(groups):
        for addons, results in groups:
            shortlist = list(filter(is_pkg, results))
            if shortlist:
                selection: Union[Defn, Tuple[()]]
                if auto:
                    pkg = shortlist[0]      # TODO: something more sophisticated
                    selection = pkg.to_defn()
                else:
                    selection = prompt_one(addons, shortlist)
                selection and (yield selection)

    def match_all():
        # Match in order of increasing heuristicitivenessitude
        for fn in (match_toc_ids,
                   match_dir_names,):
            groups = manager.run(fn(manager, (yield)))
            yield list(prompt(groups))

    leftovers = get_folders(manager)
    if not leftovers:
        click.echo('No add-ons left to reconcile.')
        return
    if not auto:
        click.echo(preamble)

    matcher = match_all()
    for _ in matcher:
        selections = matcher.send(leftovers)
        if selections and (auto or confirm('Install selected add-ons?').unsafe_ask()):
            results = manager.install(selections, replace=True)
            Report(results).generate()

        leftovers = get_folders(manager)

    if not auto and leftovers:
        click.echo()
        ctx.invoke(list_folders, exclude_own=True)


@main.command()
@click.option('--limit', '-l',
              default=5, type=click.IntRange(1, 20, clamp=True),
              help='A number to limit results to.')
@click.argument('search-terms',
                nargs=-1, required=True, callback=lambda _, __, v: ' '.join(v))
@click.pass_context
def search(ctx, limit: int, search_terms: str) -> None:
    "Search for add-ons to install."
    from .prompts import PkgChoice, checkbox, confirm

    manager = ctx.obj.m

    pkgs = manager.run(manager.search(search_terms, limit))
    if pkgs:
        choices = [PkgChoice(f'{p.name}  ({d}=={p.version})', d, pkg=p)
                   for d, p in pkgs.items()]
        selections = checkbox('Select add-ons to install', choices=choices).unsafe_ask()
        if selections and confirm('Install selected add-ons?').unsafe_ask():
            ctx.invoke(install, addons=selections)


@main.command('list')
@click.option('--export', '-e',
              default=None, type=click.Path(dir_okay=False),
              help='Export listed add-ons to CSV.')
@click.option('--format', '-f', 'output_format',
              type=click.Choice(('simple', 'detailed', 'json')),
              default='simple', show_default=True,
              help='Change the output format.')
@click.argument('addons',
                nargs=-1,
                callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@_pass_manager
def list_installed(manager, addons: Sequence[Defn], export: Optional[str], output_format: str) -> None:
    "List installed add-ons."
    from sqlalchemy import and_, or_

    def format_deps(pkg):
        deps = (Defn(pkg.origin, d.id) for d in pkg.deps)
        deps = (d.with_name(getattr(manager.get(d), 'slug', d.name)) for d in deps)
        return map(str, deps)

    def get_desc_from_toc(folders):
        toc_reader = TocReader.from_path_name(manager.config.addon_dir / pkg.folders[0].name)
        return toc_reader['Notes'].value

    pkgs = (manager.db_session.query(models.Pkg)
            .filter(or_(models.Pkg.slug.contains(d.name) if d.source == '*' else
                        and_(models.Pkg.origin == d.source,
                             or_(models.Pkg.id == d.name, models.Pkg.slug == d.name))
                        for d in addons))
            .order_by(models.Pkg.origin, models.Pkg.name)
            .all())
    if export:
        Path(export).write_text(export_to_csv(pkgs), encoding='utf-8')
    elif pkgs:
        if output_format == 'json':
            from .models import MultiPkgModel

            click.echo(MultiPkgModel.from_orm(pkgs).json(indent=2))
        elif output_format == 'detailed':
            formatter = click.HelpFormatter(max_width=99)

            for pkg in pkgs:
                with formatter.section(pkg.to_defn()):
                    formatter.write_dl((
                        ('Name', pkg.name),
                        ('Description', get_desc_from_toc(pkg) if pkg.origin == 'wowi' else pkg.description),
                        ('URL', pkg.url),
                        ('Version', pkg.version),
                        ('Date published', pkg.date_published.isoformat(' ', 'minutes')),
                        ('Folders', ', '.join(f.name for f in pkg.folders)),
                        ('Dependencies', ', '.join(format_deps(pkg)) or ''),
                        ('Options', f'{{"strategy": "{pkg.options.strategy}"}}'),))
            click.echo(formatter.getvalue(), nl=False)
        else:
            click.echo('\n'.join(str(p.to_defn()) for p in pkgs))


@main.command()
@click.option('--exclude-own', '-e',
              is_flag=True, default=False,
              help='Exclude folders managed by instawow.')
@click.option('--toc-entry', '-t', 'toc_entries',
              multiple=True,
              help='An entry to extract from the TOC.  Repeatable.')
@_pass_manager
def list_folders(manager, exclude_own: bool, toc_entries: Sequence[str]) -> None:
    "List add-on folders."
    from .matchers import get_folders

    folders = sorted(get_folders(manager, exclude_own=exclude_own))
    if folders:
        rows = [('unreconciled' if exclude_own else 'folder',
                 *(f'[{e}]' for e in toc_entries)),
                *((f.name,
                  *(f.toc_reader[e].value for e in toc_entries)) for f in folders)]
        click.echo(tabulate(rows))


@main.command(hidden=True)
@click.argument('addons', callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@click.pass_context
def info(ctx, addon: Defn) -> None:
    "Alias for `list -f detailed`."
    ctx.invoke(list_installed, addons=(addon,), output_format='detailed')


@main.command()
@click.argument('addon', callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@_pass_manager
def visit(manager, addon: Defn) -> None:
    "Open an add-on's homepage in your browser."
    pkg = manager.get_from_substr(addon)
    if pkg:
        import webbrowser
        webbrowser.open(pkg.url)
    else:
        Report({addon: E.PkgNotInstalled()}).generate_and_exit()


@main.command()
@click.argument('addon', callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@_pass_manager
def reveal(manager, addon: Defn) -> None:
    "Open an add-on folder in your file manager."
    pkg = manager.get_from_substr(addon)
    if pkg:
        import webbrowser
        webbrowser.open((manager.config.addon_dir / pkg.folders[0].name).as_uri())
    else:
        Report({addon: E.PkgNotInstalled()}).generate_and_exit()


@main.command()
def write_config() -> None:
    "Configure instawow."
    from prompt_toolkit.completion import WordCompleter
    from prompt_toolkit.shortcuts import prompt

    from .config import Config
    from .prompts import DirectoryCompleter, PydanticValidator

    addon_dir = prompt('Add-on directory: ',
                       completer=DirectoryCompleter(),
                       validator=PydanticValidator(Config, 'addon_dir'))
    game_flavour = prompt('Game flavour: ',
                          default='classic' if '_classic_' in addon_dir else 'retail',
                          completer=WordCompleter(('retail', 'classic')),
                          validator=PydanticValidator(Config, 'game_flavour'))
    config = Config(addon_dir=addon_dir, game_flavour=game_flavour).write()
    click.echo(f'Configuration written to: {config.config_file}')


@main.command()
@_pass_manager
def show_config(manager) -> None:
    "Show the active configuration."
    click.echo(manager.config.json(indent=2))


@main.group('weakauras-companion')
def _weakauras_group() -> None:
    "Manage your WeakAuras."


@_weakauras_group.command('build')
@click.option('--account', '-a',
              required=True,
              help='Your account name.  This is used to locate '
                   'the WeakAuras data file.')
@_pass_manager
def build_weakauras_companion(manager, account: str) -> None:
    "Build the WeakAuras Companion add-on."
    from .wa_updater import WaCompanionBuilder

    manager.run(WaCompanionBuilder(manager).build(account))


@_weakauras_group.command('list')
@click.option('--account', '-a',
              required=True,
              help='Your account name.  This is used to locate '
                   'the WeakAuras data file.')
@_pass_manager
def list_installed_wago_auras(manager, account: str) -> None:
    "List WeakAuras installed from Wago."
    from .wa_updater import WaCompanionBuilder

    aura_groups = WaCompanionBuilder(manager).extract_installed_auras(account)
    installed_auras = sorted((a.id, a.url, str(a.ignore_wago_update).lower())
                             for v in aura_groups.values()
                             for a in v
                             if not a.parent)
    click.echo(tabulate([('name', 'url', 'ignore updates'),
                         *installed_auras]))



from __future__ import annotations

from functools import total_ordering
from typing import TYPE_CHECKING, FrozenSet, Iterable, List, Tuple

from loguru import logger

from .models import PkgFolder
from .resolvers import Defn
from .utils import (TocReader, bucketise, cached_property, merge_intersecting_sets,
                    run_in_thread as t)

if TYPE_CHECKING:
    from .exceptions import ManagerResult
    from .manager import Manager

    _Groups = Iterable[Tuple[List[AddonFolder], List[ManagerResult]]]


_ids_to_sources = {'X-Curse-Project-ID': 'curse',
                   'X-Tukui-ProjectID': 'tukui',
                   'X-WoWI-ID': 'wowi',}


@total_ordering
class AddonFolder:
    def __init__(self, name: str, toc_reader: TocReader) -> None:
        self.name = name
        self.toc_reader = toc_reader

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.name}>'

    @cached_property
    def defns_from_toc(self) -> FrozenSet[Defn]:
        return frozenset(Defn(s, i) for s, i in
                         ((s, self.toc_reader[k].value) for k, s in _ids_to_sources.items())
                         if i)

    @cached_property
    def version(self) -> str:
        return self.toc_reader['Version',
                               'X-Packaged-Version',
                               'X-Curse-Packaged-Version'].value

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (self.__class__, str)):
            return NotImplemented
        return self.name == other

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, (self.__class__, str)):
            return NotImplemented
        return self.name < other        # type: ignore


def get_folders(manager: Manager, exclude_own: bool = True) -> FrozenSet[AddonFolder]:
    def make_addon_folder(path):
        try:
            return (path.name not in own_folders
                    and AddonFolder(path.name, TocReader.from_path_name(path)))
        except (FileNotFoundError, NotADirectoryError):
            logger.info(f'skipping {path}')

    if exclude_own:
        own_folders = {f.name for f in manager.db_session.query(PkgFolder).all()}
    else:
        own_folders = set()
    return frozenset(filter(None, map(make_addon_folder, manager.config.addon_dir.iterdir())))


async def match_toc_ids(manager: Manager, leftovers: FrozenSet[AddonFolder]) -> _Groups:
    "Attempt to match add-ons from source IDs contained in TOC files."
    def keyer(value):
        return next(d for d in defns if value.defns_from_toc & d)

    matches = [a for a in sorted(leftovers) if a.defns_from_toc]
    defns = list(merge_intersecting_sets(a.defns_from_toc for a in matches))
    results = await manager.resolve(list(frozenset.union(*defns, frozenset())))      # type: ignore
    return [(f, [results[d] for d in b]) for b, f in bucketise(matches, keyer).items()]


async def match_dir_names(manager: Manager, leftovers: FrozenSet[AddonFolder]) -> _Groups:
    "Attempt to match folders against the CurseForge and WoWInterface catalogues."
    async def fetch_combined_folders():
        url = ('https://raw.githubusercontent.com/layday/instascrape/data/'
               'combined-folders.compact.json')   # v1
        async with manager.web_client.get(url) as response:
            return await response.json(content_type=None)

    def keyer(value):
        return next(f for f in folders if value[0] & f)

    # We can't use an intersection here because it's not guaranteed to return ``AddonFolder``s
    matches = [(frozenset(filter(f.__contains__, leftovers)), Defn(*d))
               for d, c, f in await fetch_combined_folders()
               if manager.config.game_flavour in c
               and frozenset(f) <= leftovers]
    folders = list(merge_intersecting_sets(f for f, _ in matches))
    results = await manager.resolve(list({d for _, d in matches}))
    return [(sorted(f), [results[d] for _, d in b]) for f, b in bucketise(matches, keyer).items()]



from alembic.command import (branches, current, downgrade, edit, heads, history, init,
                             list_templates, merge, revision, show, stamp, upgrade)
from alembic.config import Config as AlembicConfig


def make_config(url: str) -> AlembicConfig:
    config = AlembicConfig()
    config.set_main_option('script_location', 'instawow:migrations')
    config.set_main_option('sqlalchemy.url', url)
    return config



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
import contextvars as cv
from functools import partial
from itertools import filterfalse, repeat, starmap
from pathlib import Path, PurePath
import posixpath
from shutil import copy, move
from tempfile import NamedTemporaryFile, mkdtemp
from typing import (TYPE_CHECKING, Any, AsyncContextManager as ACM, AsyncGenerator, Awaitable,
                    Callable, Dict, Hashable, Iterable, List, NoReturn, Optional, Sequence, Set,
                    Tuple)

from loguru import logger

from . import DB_REVISION, exceptions as E
from .models import Pkg, PkgFolder, is_pkg
from .resolvers import CurseResolver, Defn, InstawowResolver, TukuiResolver, WowiResolver
from .utils import (bucketise, cached_property, dict_merge as merge, gather, make_progress_bar,
                    run_in_thread as t, shasum)

if TYPE_CHECKING:
    import aiohttp
    from prompt_toolkit.shortcuts import ProgressBar
    from sqlalchemy.orm import scoped_session
    from .config import Config


USER_AGENT = 'instawow (https://github.com/layday/instawow)'

_web_client: cv.ContextVar[aiohttp.ClientSession] = cv.ContextVar('_web_client')

AsyncNamedTemporaryFile = t(NamedTemporaryFile)
async_mkdtemp = t(mkdtemp)
async_copy = t(copy)
async_move = t(move)


@asynccontextmanager
async def open_temp_writer() -> AsyncGenerator[Tuple[Path, Callable], None]:
    fh = await AsyncNamedTemporaryFile(delete=False)
    path = Path(fh.name)
    try:
        yield (path, t(fh.write))
    except BaseException:
        await t(fh.close)()
        await t(path.unlink)()
        raise
    else:
        await t(fh.close)()


async def trash(paths: Sequence[Path], parent_dir: PurePath, *, missing_ok: bool = False) -> None:
    dst = await async_mkdtemp(dir=parent_dir, prefix=paths[0].name + '-')
    for path in map(str, paths):    # https://bugs.python.org/issue32689
        try:
            await async_move(path, dst)
        except (FileNotFoundError if missing_ok else ()):  # type: ignore  # https://github.com/python/mypy/issues/7356
            logger.opt(exception=True).info('source missing')


# macOS 'resource forks' are sometimes included in download zips - these
# violate our one add-on per folder contract-thing and will be omitted
_zip_excludes = {'__MACOSX'}


def find_base_dirs(names: Sequence[str]) -> Set[str]:
    return {n for n in (posixpath.dirname(n) for n in names)
            if n and posixpath.sep not in n} - _zip_excludes


def should_extract(base_dirs: Set[str]) -> Callable[[str], bool]:
    def is_member(name):
        head, sep, _ = name.partition(posixpath.sep)
        return sep and head in base_dirs

    return is_member


@asynccontextmanager
async def acquire_archive(path: PurePath) -> AsyncGenerator[Tuple[List[str], Callable], None]:
    from zipfile import ZipFile

    def extract(parent: Path) -> None:
        conflicts = base_dirs & {f.name for f in parent.iterdir()}
        if conflicts:
            raise E.PkgConflictsWithForeign(conflicts)
        else:
            members = filter(should_extract(base_dirs), names)
            archive.extractall(parent, members=members)

    archive = await t(ZipFile)(path)
    try:
        names = archive.namelist()
        base_dirs = find_base_dirs(names)
        yield (sorted(base_dirs), t(extract))
    finally:
        await t(archive.close)()


async def download_archive(manager: Manager, pkg: Pkg, *, chunk_size: int = 4096) -> ACM:
    url = pkg.download_url
    dst = manager.config.cache_dir / shasum(pkg.origin, pkg.id, pkg.file_id)

    if await t(dst.exists)():
        pass
    elif url.startswith('file://'):
        from urllib.parse import unquote

        await async_copy(unquote(url[7:]), dst)
    else:
        web_client = _web_client.get()
        async with web_client.get(url, trace_request_ctx={'show_progress': True}) as response, \
                open_temp_writer() as (temp_path, write):
            async for chunk in response.content.iter_chunked(chunk_size):
                await write(chunk)

        await async_move(str(temp_path), dst)
    return acquire_archive(dst)


def prepare_db_session(config: Config) -> scoped_session:
    from sqlalchemy import create_engine
    from sqlalchemy.orm import sessionmaker, scoped_session
    from .models import ModelBase, should_migrate

    db_path = config.config_dir / 'db.sqlite'
    db_url = f'sqlite:///{db_path}'
    # We need to distinguish between newly-created databases and
    # databases predating alembic in instawow - attempting to migrate a
    # new database will throw an error
    db_exists = db_path.exists()

    engine = create_engine(db_url)
    if should_migrate(engine, DB_REVISION):
        from .migrations import make_config, stamp, upgrade

        alembic_config = make_config(db_url)
        if db_exists:
            logger.info(f'migrating database to {DB_REVISION}')
            upgrade(alembic_config, DB_REVISION)
        else:
            ModelBase.metadata.create_all(engine)
            logger.info(f'stamping database with {DB_REVISION}')
            stamp(alembic_config, DB_REVISION)

    return scoped_session(sessionmaker(bind=engine))


def init_web_client(**kwargs: Any) -> aiohttp.ClientSession:
    from aiohttp import ClientSession, TCPConnector

    kwargs = {'connector': TCPConnector(force_close=True, limit_per_host=10),
              'headers': {'User-Agent': USER_AGENT},
              'trust_env': True,    # Respect http_proxy env var
              **kwargs}
    return ClientSession(**kwargs)


@object.__new__
class _DummyResolver:
    async def synchronise(self) -> _DummyResolver:
        return self

    async def resolve(self, defns: List[Defn], **kwargs: Any) -> Dict[Defn, E.PkgOriginInvalid]:
        return dict(zip(defns, repeat(E.PkgOriginInvalid())))


class _ResolverDict(dict):
    RESOLVERS = {CurseResolver, WowiResolver, TukuiResolver, InstawowResolver}

    def __init__(self, manager: Manager) -> None:
        super().__init__((r.source, r(manager=manager)) for r in self.RESOLVERS)

    def __missing__(self, key: Hashable) -> _DummyResolver:
        return _DummyResolver       # type: ignore      # unreported?


async def _error_out(error: E.ManagerError) -> NoReturn:
    raise error


class Manager:
    def __init__(self, config: Config, db_session: scoped_session) -> None:
        self.config = config
        self.db_session = db_session
        self.resolvers = _ResolverDict(self)

    @property
    def web_client(self) -> aiohttp.ClientSession:
        return _web_client.get()

    @web_client.setter
    def web_client(self, value: aiohttp.ClientSession) -> None:
        _web_client.set(value)

    def get(self, defn: Defn) -> Optional[Pkg]:
        return (self.db_session.query(Pkg)
                .filter(Pkg.origin == defn.source,
                        (Pkg.id == defn.name) | (Pkg.slug == defn.name)).first())

    def get_from_substr(self, defn: Defn) -> Optional[Pkg]:
        pkg = self.get(defn)
        if not pkg:
            pkg = (self.db_session.query(Pkg)
                   .filter(Pkg.slug.contains(defn.name)).order_by(Pkg.name).first())
        return pkg

    async def _resolve_deps(self, results: Iterable[Any]) -> Dict[Defn, Any]:
        """Resolve package dependencies.

        The resolver will not follow dependencies
        more than one level deep.  This is to avoid unnecessary
        complexity for something that I would never expect to
        encounter in the wild.
        """
        pkgs = list(filter(is_pkg, results))
        dep_defns = list(filterfalse({(p.origin, p.id) for p in pkgs}.__contains__,
                                     # Using a dict to maintain dep appearance order
                                     {(p.origin, d.id): ... for p in pkgs for d in p.deps}))
        if not dep_defns:
            return {}

        deps = await self.resolve(list(starmap(Defn, dep_defns)))
        pretty_deps = {d.with_name(r.slug) if is_pkg(r) else d: r for d, r in deps.items()}
        return pretty_deps

    async def resolve(self, defns: Sequence[Defn], with_deps: bool = False) -> Dict[Defn, Any]:
        "Resolve definitions into packages."
        async def get_results(source, defns):
            resolver = await self.resolvers[source].synchronise()
            return await resolver.resolve(defns)

        defns_by_source = bucketise(defns, key=lambda v: v.source)
        results = await gather(get_results(s, b) for s, b in defns_by_source.items())
        results_by_defn = merge(dict.fromkeys(defns), *results)
        if with_deps:
            results_by_defn.update(await self._resolve_deps(results_by_defn.values()))
        return results_by_defn

    async def search(self, search_terms: str, limit: int) -> Dict[Defn, Pkg]:
        "Search the combined names catalogue for packages."
        from fuzzywuzzy import fuzz, process
        from .resolvers import _FileCacheMixin as cache

        url = ('https://raw.githubusercontent.com/layday/instascrape/data/'
               'combined-names-v2.compact.json')   # v2
        combined_names = await cache._cache_json_response(self, url, 8, 'hours')
        defns_for_names = bucketise(((n, Defn(*d)) for n, d, f in combined_names
                                     if self.config.game_flavour in f),
                                    key=lambda v: v[0])

        matches = process.extract(search_terms, defns_for_names.keys(),
                                  limit=limit, scorer=fuzz.WRatio)
        defns = [d
                 for m, _ in matches
                 for _, d in defns_for_names[m]]
        results = await self.resolve(defns)
        pkgs_by_defn = {d.with_name(r.slug): r for d, r in results.items() if is_pkg(r)}
        return pkgs_by_defn

    async def install_one(self, pkg: Pkg, archive: ACM, replace: bool) -> E.PkgInstalled:
        async with archive as (folders, extract):
            conflicts = (self.db_session.query(Pkg).join(Pkg.folders)
                         .filter(PkgFolder.name.in_(folders)).all())
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts)
            if replace:
                await trash([self.config.addon_dir / f for f in folders], self.config.temp_dir)
            await extract(self.config.addon_dir)

        pkg.folders = [PkgFolder(name=f) for f in folders]
        self.db_session.add(pkg)
        self.db_session.commit()
        return E.PkgInstalled(pkg)

    async def prep_install(self, defns: Sequence[Defn], replace: bool) -> Dict[Defn, Callable]:
        "Retrieve packages to install."
        prelim_results = await self.resolve([d for d in defns if not self.get(d)],
                                            with_deps=True)
        # Weed out installed deps - this isn't super efficient but
        # avoids having to deal with local state in resolver
        results = {d: r for d, r in prelim_results.items() if not self.get(d)}
        installables = {(d, r): download_archive(self, r) for d, r in results.items() if is_pkg(r)}
        archives = await gather(installables.values())

        return merge(dict.fromkeys(defns, partial(_error_out, E.PkgAlreadyInstalled())),
                     {d: partial(_error_out, r) for d, r in results.items()},
                     {d: partial(self.install_one, p, a, replace)
                      for (d, p), a in zip(installables, archives)})

    async def update_one(self, old_pkg: Pkg, pkg: Pkg, archive: ACM) -> E.PkgUpdated:
        async with archive as (folders, extract):
            conflicts = (self.db_session.query(Pkg).join(Pkg.folders)
                         .filter(PkgFolder.pkg_origin != pkg.origin, PkgFolder.pkg_id != pkg.id)
                         .filter(PkgFolder.name.in_(folders)).all())
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts)

            await self.remove_one(old_pkg)
            await extract(self.config.addon_dir)

        pkg.folders = [PkgFolder(name=f) for f in folders]
        self.db_session.add(pkg)
        self.db_session.commit()
        return E.PkgUpdated(old_pkg, pkg)

    async def prep_update(self, defns: Sequence[Defn]) -> Dict[Defn, Callable]:
        "Retrieve packages to update."
        # Rebuild ``Defn`` with strategy from package for ``Defn``s
        # with installed packages
        checked_defns = {c.with_strategy(p.to_defn().strategy) if p else c: p
                         for c, p in ((d, self.get(d)) for d in defns)}
        # Results can contain errors
        # installables are those results which are packages
        # and updatables are packages with updates
        results = await self.resolve([d for d, p in checked_defns.items() if p])
        installables = {d: r for d, r in results.items() if is_pkg(r)}
        updatables = {(d, checked_defns[d], p): download_archive(self, p)
                      for d, p in installables.items()
                      if p.file_id != checked_defns[d].file_id}
        archives = await gather(updatables.values())

        return merge(dict.fromkeys(checked_defns, partial(_error_out, E.PkgNotInstalled())),
                     {d: partial(_error_out, r) for d, r in results.items()},
                     {d: partial(_error_out, E.PkgUpToDate()) for d in installables},
                     {d: partial(self.update_one, *p, a)
                      for (d, *p), a in zip(updatables, archives)})

    async def remove_one(self, pkg: Pkg) -> E.PkgRemoved:
        await trash([self.config.addon_dir / f.name for f in pkg.folders],
                    parent_dir=self.config.temp_dir, missing_ok=True)
        self.db_session.delete(pkg)
        self.db_session.commit()
        return E.PkgRemoved(pkg)

    async def prep_remove(self, defns: Sequence[Defn]) -> Dict[Defn, Callable]:
        "Prepare packages to remove."
        pkgs_by_defn = ((d, self.get(d)) for d in defns)
        return merge(dict.fromkeys(defns, partial(_error_out, E.PkgNotInstalled())),
                     {d: partial(self.remove_one, p) for d, p in pkgs_by_defn if p})


_tick_interval = .1
_tickers: cv.ContextVar[Set[asyncio.Task]] = cv.ContextVar('_tickers', default=set())


@asynccontextmanager
async def cancel_tickers() -> AsyncGenerator[None, None]:
    try:
        yield
    finally:
        for ticker in _tickers.get():
            ticker.cancel()


def init_cli_web_client(*, make_bar: ProgressBar) -> aiohttp.ClientSession:
    from cgi import parse_header
    from aiohttp import TraceConfig

    def extract_filename(params) -> str:
        _, cd_params = parse_header(params.response.headers.get('Content-Disposition', ''))
        filename = cd_params.get('filename') or params.response.url.name
        return filename

    async def do_on_request_end(session, request_ctx, params) -> None:
        ctx = request_ctx.trace_request_ctx
        if ctx and ctx.get('show_progress'):
            bar = make_bar(label=ctx.get('label') or f'Downloading {extract_filename(params)}',
                           total=params.response.content_length or 0)

            async def ticker() -> None:
                try:
                    content = params.response.content
                    while not content.is_eof():
                        bar.current = content._cursor
                        await asyncio.sleep(_tick_interval)
                finally:
                    bar.progress_bar.counters.remove(bar)

            tickers = _tickers.get()
            tickers.add(asyncio.create_task(ticker()))

    trace_config = TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()
    return init_web_client(trace_configs=[trace_config])


class CliManager(Manager):
    def __init__(self, config: Config, db_session: scoped_session,
                 progress_bar_factory: Callable = make_progress_bar) -> None:
        super().__init__(config, db_session)
        self.progress_bar_factory = progress_bar_factory

    @cached_property
    def progress_bar(self) -> Any:
        return self.progress_bar_factory()

    def run(self, awaitable: Awaitable) -> Any:
        async def run():
            async with init_cli_web_client(make_bar=self.progress_bar) as self.web_client, \
                    cancel_tickers():
                return await awaitable

        with self.progress_bar:
            return asyncio.run(run())

    def _prepprocess(self, prepper: Callable, *args: Any, **kwargs: Any) -> Dict[Defn, E.ManagerResult]:
        async def intercept(fn):
            try:
                return await fn()
            except E.ManagerError as error:
                return error
            except Exception as error:
                logger.exception('internal error')
                return E.InternalError(error)

        async def process():
            coros_by_defn = await prepper(*args, **kwargs)
            return {d: await intercept(c) for d, c in coros_by_defn.items()}

        return self.run(process())

    @property
    def install(self) -> Callable:
        return partial(self._prepprocess, self.prep_install)

    @property
    def update(self) -> Callable:
        return partial(self._prepprocess, self.prep_update)

    @property
    def remove(self) -> Callable:
        return partial(self._prepprocess, self.prep_remove)



from pathlib import Path

from setuptools import find_packages, setup


metadata = {
    'name': 'instawow',
    'url': 'http://github.com/layday/instawow',
    'author': 'layday',
    'author_email': 'layday@protonmail.com',
    'license': 'GPL-3.0-or-later',
    'description': 'A CLI for managing World of Warcraft add-ons.',
    'long_description': Path('README.rst').read_text(encoding='utf-8'),}

options = {
    'entry_points': {'console_scripts': ['instawow = instawow.cli:main']},
    'install_requires': Path('requirements.txt').read_text(),
    'python_requires': '~=3.7',
    'include_package_data': True,
    'packages': find_packages(),
    'use_scm_version': True,}

setup(**metadata, **options)
